/*
 * Copyright 2003,2004,2006 Red Hat, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of the
 * GNU Lesser General Public License, in which case the provisions of the
 * LGPL are required INSTEAD OF the above restrictions.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "../config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef HAVE_SECURITY_PAM_MODULES_H
#include <security/pam_modules.h>
#endif

#include <krb5.h>
#ifdef USE_KRB4
#include KRB4_DES_H
#include KRB4_KRB_H
#ifdef KRB4_KRB_ERR_H
#include KRB4_KRB_ERR_H
#endif
#endif

#include "init.h"
#include "log.h"
#include "stash.h"
#include "storetmp.h"
#include "userinfo.h"
#include "v5.h"
#include "xstr.h"

#ident "$Id: stash.c,v 1.8.2.1 2006/01/18 22:38:15 nalin Exp $"

static int
_pam_krb5_get_data_stash(pam_handle_t *pamh, const char *key,
			 struct _pam_krb5_stash **stash)
{
	return pam_get_data(pamh, key, (PAM_KRB5_MAYBE_CONST void**) stash);
}

/* Clean up a stash.  This includes freeing any dynamically-allocated bits and
 * then freeing the stash itself. */
static void
_pam_krb5_stash_cleanup(pam_handle_t *pamh, void *data, int error)
{
	struct _pam_krb5_stash *stash = data;
	krb5_free_cred_contents(stash->v5ctx, &stash->v5creds);
	free(stash->key);
	if (stash->v5file != NULL) {
		xstrfree(stash->v5file);
	}
#ifdef USE_KRB4
	if (stash->v4file != NULL) {
		xstrfree(stash->v4file);
	}
#endif
	memset(stash, 0, sizeof(struct _pam_krb5_stash));
	free(stash);
}

/* Check for KRB5CCNAME and KRBTKFILE in the PAM environment.  If either
 * exists, incorporate contents of the named ccache/tktfiles into the stash. */
static void
_pam_krb5_stash_external_read(pam_handle_t *pamh, struct _pam_krb5_stash *stash,
			      struct _pam_krb5_options *options)
{
	krb5_context ctx;
	krb5_ccache ccache;
	krb5_cc_cursor cursor;
	int i;
	const char *ccname;
	char *unparsed;

	/* Read a TGT from $KRB5CCNAME. */
	ccname = pam_getenv(pamh, "KRB5CCNAME");
	if ((ccname != NULL) && (strlen(ccname) > 0)) {
		if (options->debug) {
			debug("KRB5CCNAME is set to \"%s\"", ccname);
		}
		if (stash->v5ctx != NULL) {
			ctx = stash->v5ctx;
		} else {
			if (_pam_krb5_init_ctx(&ctx, 0, NULL) != PAM_SUCCESS) {
				warn("error initializing kerberos");
				return;
			}
		}
		ccache = NULL;
		i = krb5_cc_resolve(ctx, ccname, &ccache);
		if (i != 0) {
			warn("error opening ccache \"%s\", ignoring", ccname);
		} else {
			cursor = NULL;
			if (krb5_cc_start_seq_get(ctx, ccache, &cursor) == 0) {
				memset(&stash->v5creds, 0, sizeof(stash->v5creds));
				while (krb5_cc_next_cred(ctx, ccache, &cursor,
							 &stash->v5creds) == 0) {
					unparsed = NULL;
					i = krb5_unparse_name(ctx,
							      stash->v5creds.server,
							      &unparsed);
					if ((i == 0) && (unparsed != NULL)) {
						i = strcspn(unparsed,
							    PAM_KRB5_PRINCIPAL_COMPONENT_SEPARATORS);
						if (strncmp(unparsed,
							    KRB5_TGS_NAME,
							    KRB5_TGS_NAME_SIZE) == 0) {
							if (options->debug) {
								debug("using credential for \"%s\" as a v5 TGT", unparsed);
							}
							v5_free_unparsed_name(ctx, unparsed);
							unparsed = NULL;
							stash->v5attempted = 1;
							stash->v5result = 0;
							break;
						}
						if (options->debug) {
							debug("not using credential for \"%s\" as a v5 TGT", unparsed);
						}
						v5_free_unparsed_name(ctx, unparsed);
						unparsed = NULL;
					}
					krb5_free_cred_contents(ctx, &stash->v5creds);
					memset(&stash->v5creds, 0, sizeof(stash->v5creds));
				}
				krb5_cc_end_seq_get(ctx, ccache, &cursor);
			}
			krb5_cc_close(ctx, ccache);
		}
		if (ctx != stash->v5ctx) {
			krb5_free_context(ctx);
		}
	}

#if 0
#ifdef USE_KRB4
	const char *v4tktname; /* FIXME: not available before C99! */
	/* Read a TGT from $KRBTKFILE. */
	v4tktname = pam_getenv(pamh, "KRBTKFILE");
	if ((v4tktname != NULL) && (strlen(v4tktname) > 0) &&
	    (stash->v4present == 0)) {
		char name[ANAME_SZ + 1], instance[INST_SZ + 1],
		     realm[REALM_SZ + 1];

		if (tf_init(pam_getenv(pamh, "KRBTKFILE"), R_TKT_FILE) == 0) {
			if ((tf_get_pname(name) == 0) &&
			    (tf_get_pinst(instance) == 0)) {
				while (tf_get_cred(&stash->v4creds) == 0) {
					if (strncmp(stash->v4creds.service,
						    KRB5_TGS_NAME,
						    KRB5_TGS_NAME_SIZE) == 0) {
	    					stash->v4present = 1;
						break;
					}
				}
			}
			tf_close();
		}
	}
#endif
#endif
}

/* Get the stash of lookaside data we keep about this user.  If we don't
 * already have one, we need to create it.  We use a data name which includes
 * the principal name to allow checks within multiple realms to work, and we
 * store the key in the stash because older versions of libpam stored the
 * pointer instead of making their own copy of the key, which could lead to
 * crashes if we then deallocated the string. */
struct _pam_krb5_stash *
_pam_krb5_stash_get(pam_handle_t *pamh, struct _pam_krb5_user_info *info,
		    struct _pam_krb5_options *options)
{
	struct _pam_krb5_stash *stash;
	char *key;

	key = malloc(strlen("_pam_krb5_stash_") +
		     strlen(info->unparsed_name) +
		     1);
	if (key == NULL) {
		return NULL;
	}
	sprintf(key, "_pam_krb5_stash_%s", info->unparsed_name);

	stash = NULL;
	if ((_pam_krb5_get_data_stash(pamh, key, &stash) == PAM_SUCCESS) &&
	    (stash != NULL)) {
		free(key);
		if (options->external && (stash->v5attempted == 0)) {
			_pam_krb5_stash_external_read(pamh, stash, options);
		}
		return stash;
	}

	stash = malloc(sizeof(struct _pam_krb5_stash));
	if (stash == NULL) {
		free(key);
		return NULL;
	}
	memset(stash, 0, sizeof(struct _pam_krb5_stash));

	stash->key = key;
	stash->v5ctx = NULL;
	stash->v5attempted = 0;
	stash->v5result = KRB5KRB_ERR_GENERIC;
	stash->v5file = NULL;
	memset(&stash->v5creds, 0, sizeof(stash->v5creds));
#ifdef USE_KRB4
	stash->v4present = 0;
	memset(&stash->v4creds, 0, sizeof(stash->v4creds));
	stash->v4file = NULL;
#endif
	stash->afspag = 0;
	if (options->external && (stash->v5attempted == 0)) {
		_pam_krb5_stash_external_read(pamh, stash, options);
	}
	pam_set_data(pamh, key, stash, _pam_krb5_stash_cleanup);

	return stash;
}

static void
_pam_krb5_stash_clone(char **stored_file, uid_t uid, gid_t gid)
{
	char *filename;
	size_t length;
	if (*stored_file != NULL) {
		filename = strdup(*stored_file);
		length = strlen(filename);
		if (length > 6) {
			strcpy(filename + length - 6, "XXXXXX");
			if (_pam_krb5_storetmp_file(*stored_file,
						    filename, uid, gid,
						    filename,
						    length + 1) == 0) {
				unlink(*stored_file);
				xstrfree(*stored_file);
				*stored_file = filename;
			}
		}
		if (*stored_file != filename) {
			free(filename);
		}
	}
}

void
_pam_krb5_stash_clone_v5(struct _pam_krb5_stash *stash, uid_t uid, gid_t gid)
{
	_pam_krb5_stash_clone(&stash->v5file, uid, gid);
}

#ifdef USE_KRB4
void
_pam_krb5_stash_clone_v4(struct _pam_krb5_stash *stash, uid_t uid, gid_t gid)
{
	_pam_krb5_stash_clone(&stash->v4file, uid, gid);
}
#else
void
_pam_krb5_stash_clone_v4(struct _pam_krb5_stash *stash, uid_t uid, gid_t gid)
{
}
#endif

static int
_pam_krb5_stash_clean(char **stored_file)
{
	if (_pam_krb5_storetmp_delete(*stored_file) == 0) {
		xstrfree(*stored_file);
		*stored_file = NULL;
		return 0;
	} else {
		if (unlink(*stored_file) == 0) {
			xstrfree(*stored_file);
			*stored_file = NULL;
			return 0;
		}
	}
	return -1;
}

#ifdef USE_KRB4
int
_pam_krb5_stash_clean_v4(struct _pam_krb5_stash *stash)
{
	return _pam_krb5_stash_clean(&stash->v4file);
}
#else
int
_pam_krb5_stash_clean_v4(struct _pam_krb5_stash *stash)
{
	return 0;
}
#endif

int
_pam_krb5_stash_clean_v5(struct _pam_krb5_stash *stash)
{
	return _pam_krb5_stash_clean(&stash->v5file);
}
